#include <string.h>
#include "db_buffer.h"
#include "db_vm.h"

typedef struct {
    int lineNumber;
    int length;
    char text[1];
} Line;

static uint8_t *buffer;
static uint8_t *bufferMax;
static uint8_t *bufferTop;
static Line *current;

static uint8_t *data;
static uint8_t *dataPtr;
static uint8_t *dataTop;

static int FindLineN(int lineNumber, Line **pLine);

int BufInit(System *sys, size_t editSize, size_t scratchSize)
{
    if (!(buffer = AllocateFreeSpace(sys, editSize))
    ||  !(data = AllocateFreeSpace(sys, scratchSize)))
        return VMFALSE;
    bufferMax = buffer + editSize;
    dataTop = data + scratchSize;
    EditBufReset();
    ScratchRewind();
    return VMTRUE;
}

void EditBufReset(void)
{
    bufferTop = buffer;
    current = (Line *)buffer;
}

int EditBufAddLineN(int lineNumber, const char *text)
{
    int newLength = sizeof(Line) + strlen(text);
    Line *line, *next;
    int spaceNeeded;

    /* make sure the length is a multiple of the word size */
    newLength = (newLength + ALIGN_MASK) & ~ALIGN_MASK;

    /* replace an existing line */
    if (FindLineN(lineNumber, &line)) {
        next = (Line *)((uint8_t *)line + line->length);
        spaceNeeded = newLength - line->length;
    }

    /* insert a new line */
    else {
        next = line;
        spaceNeeded = newLength;
    }

    /* make sure there is enough space */
    if (bufferTop + spaceNeeded > bufferMax)
        return VMFALSE;

    /* make space for the new line */
    if ((uint8_t *)next < bufferTop)
        memmove((uint8_t *)next + spaceNeeded, next, bufferTop - (uint8_t *)next);
    bufferTop += spaceNeeded;

    /* insert the new line */
    if (newLength > 0) {
        line->lineNumber = lineNumber;
        line->length = newLength;
        strcpy(line->text, text);
    }

    /* return successfully */
    return VMTRUE;
}

int EditBufDeleteLineN(int lineNumber)
{
    Line *line, *next;
    int spaceFreed;

    /* find the line to delete */
    if (!FindLineN(lineNumber, &line))
        return VMFALSE;

    /* get a pointer to the next line */
    next = (Line *)((uint8_t *)line + line->length);
    spaceFreed = line->length;

    /* remove the line to be deleted */
    if ((uint8_t *)next < bufferTop)
        memmove(line, next, bufferTop - (uint8_t *)next);
    bufferTop -= spaceFreed;

    /* return successfully */
    return VMTRUE;
}

int EditBufSeekN(int lineNumber)
{
    /* if the line number is zero start at the first line */
    if (lineNumber == 0)
        current = (Line *)buffer;

    /* otherwise, start at the specified line */
    else if (!FindLineN(lineNumber, &current))
        return VMFALSE;

    /* return successfully */
    return VMTRUE;
}

int EditBufGetLine(int *pLineNumber, char *text)
{
    /* check for the end of the buffer */
    if ((uint8_t *)current >= bufferTop)
        return VMFALSE;

    /* get the current line */
    *pLineNumber = current->lineNumber;
    strcpy(text, current->text);

    /* move ahead to the next line */
    current = (Line *)((char *)current + current->length);

    /* return successfully */
    return VMTRUE;
}

static int FindLineN(int lineNumber, Line **pLine)
{
    uint8_t *p = buffer;
    while (p < bufferTop) {
        Line *line = (Line *)p;
        if (lineNumber <= line->lineNumber) {
            *pLine = line;
            return lineNumber == line->lineNumber;
        }
        p = p + line->length;
    }
    *pLine = (Line *)p;
    return VMFALSE;
}

void ScratchRewind(void)
{
    dataPtr = data;
}

int ScratchWrite(const uint8_t *buf, int size)
{
    if (dataPtr + size > dataTop)
        return VMFALSE;
    memcpy(dataPtr, buf, size);
    dataPtr += size;
    return VMTRUE;
}

int ScratchRead(uint8_t *buf, int size)
{
    if (dataPtr + size > dataTop)
        return VMFALSE;
    memcpy(buf, dataPtr, size);
    dataPtr += size;
    return VMTRUE;
}

