/* db_vmfcn.c - intrinsic functions
 *
 * Copyright (c) 2011 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "db_vm.h"

static void fcn_abs(Interpreter *i);
static void fcn_rnd(Interpreter *i);
static void fcn_left(Interpreter *i);
static void fcn_right(Interpreter *i);
static void fcn_mid(Interpreter *i);
static void fcn_chr(Interpreter *i);
static void fcn_str(Interpreter *i);
static void fcn_asc(Interpreter *i);
static void fcn_len(Interpreter *i);

/* this table must be in the same order as the FN_xxx macros */
IntrinsicFcn * FLASH_SPACE Intrinsics[] = {
    fcn_abs,
	fcn_rnd,
    fcn_left,
    fcn_right,
    fcn_mid,
    fcn_chr,
    fcn_str,
    fcn_asc,
    fcn_len
};

int IntrinsicCount = sizeof(Intrinsics) / sizeof(IntrinsicFcn *);

/* fcn_abs - ABS(n): return the absolute value of a number */
static void fcn_abs(Interpreter *i)
{
    i->tos = abs(Pop(i));
}

/* fcn_rnd - RND(n): return a random number between 0 and n-1 */
static void fcn_rnd(Interpreter *i)
{
    i->tos = rand() % Pop(i);
}

/* fcn_left - LEFT$(str$, n): return the leftmost n characters of a string */
static void fcn_left(Interpreter *i)
{
    ObjectHdr *hdr, *val;
    VMVALUE hStr, count;
    VMUVALUE size;
    
    /* get the string */
    hStr = i->sp[-1];
    CheckObjNumber(i, hStr);
    hdr = GetObjHdr(i->image, hStr);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    size = GetVectorSize(hdr);
    
    /* get the character count */
    count = i->sp[0];
    
    /* determine the length of the result */
    if (count > size)
        count = size;
    else if (count < 0)
        Abort(i, "count can not be negative");
    
    /* allocate the result string */
    if (!(i->tos = StringAlloc(i->image, count)))
        Abort(i, "insufficient heap space");
    val = GetObjHdr(i->image, i->tos);
    
    /* copy the substring */
    memcpy(GetBVectorBase(val), GetBVectorBase(hdr), count);
    
    Drop(i, 2);
}

/* fcn_right - RIGHT$(str$, n): return the rightmost n characters of a string */
static void fcn_right(Interpreter *i)
{
    ObjectHdr *hdr, *val;
    VMVALUE hStr, count;
    VMUVALUE size;
    
    /* get the string */
    hStr = i->sp[-1];
    CheckObjNumber(i, hStr);
    hdr = GetObjHdr(i->image, hStr);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    size = GetVectorSize(hdr);
    
    /* get the character count */
    count = i->sp[0];
    
    /* determine the length of the result */
    if (count > size)
        count = size;
    else if (count < 0)
        Abort(i, "count can not be negative");
    
    /* allocate the result string */
    if (!(i->tos = StringAlloc(i->image, count)))
        Abort(i, "insufficient heap space");
    val = GetObjHdr(i->image, i->tos);
    
    /* copy the substring */
    memcpy(GetBVectorBase(val), GetBVectorBase(hdr) + size - count, count);
    
    Drop(i, 2);
}

/* fcn_mid - MID$(str$, start, n): return n characters from the middle of a string */
static void fcn_mid(Interpreter *i)
{
    ObjectHdr *hdr, *val;
    VMVALUE hStr, start, count;
    VMUVALUE size;
    
    /* get the string */
    hStr = i->sp[-2];
    CheckObjNumber(i, hStr);
    hdr = GetObjHdr(i->image, hStr);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    size = GetVectorSize(hdr);
    
    /* get the starting position */
    start = i->sp[-1];
    
    /* get the character count */
    count = i->sp[0];
    
    /* check the starting position and count */
    if (start < 0 || start >= size)
        Abort(i, "string index error");
    else if (start + count > size)
        count = size - start;
    
    /* allocate the result string */
    if (!(i->tos = StringAlloc(i->image, count)))
        Abort(i, "insufficient heap space");
    val = GetObjHdr(i->image, i->tos);
    
    /* copy the substring */
    memcpy(GetBVectorBase(val), GetBVectorBase(hdr) + start, count);
    
    Drop(i, 3);
}

/* fcn_chr - CHR$(n): return a one character string with the specified character code */
static void fcn_chr(Interpreter *i)
{
    ObjectHdr *val;
    i->tos = StringAlloc(i->image, 1);
    val = GetObjHdr(i->image, i->tos);
    GetBVectorBase(val)[0] = i->sp[0];
    Drop(i, 1);
}

/* fcn_str - STR$(n): return n converted to a string */
static void fcn_str(Interpreter *i)
{
    ObjectHdr *val;
    VMUVALUE len;
    char buf[16];
    sprintf(buf, "%d", i->sp[0]);
    len = (VMUVALUE)strlen(buf);
    if (!(i->tos = StringAlloc(i->image, len)))
        Abort(i, "insufficient heap space");
    val = GetObjHdr(i->image, i->tos);
    memcpy(GetBVectorBase(val), (uint8_t *)buf, len);
    Drop(i, 1);
}

/* fcn_asc - ASC(str$): return the character code of the first character of a string */
static void fcn_asc(Interpreter *i)
{
    VMVALUE hStr = i->sp[0];
    ObjectHdr *hdr;
    
    /* get the first string pointer and length */
    CheckObjNumber(i, hStr);
    hdr = GetObjHdr(i->image, hStr);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    
    /* return the first character of the string or -1 if the string is empty */
    i->tos = (GetVectorSize(hdr) > 0 ?  GetBVectorBase(hdr)[0] : -1);
    
    Drop(i, 1);
}

/* fcn_len - LEN(str$): return length of a string */
static void fcn_len(Interpreter *i)
{
    VMVALUE hStr = i->sp[0];
    ObjectHdr *hdr;
    
    /* get the first string pointer and length */
    CheckObjNumber(i, hStr);
    hdr = GetObjHdr(i->image, hStr);
    CheckObjType(i, hdr, PROTO_BVECTOR);
    
    /* return the length of the string */
    i->tos = (VMVALUE)GetVectorSize(hdr);
    
    Drop(i, 1);
}
